
#include "meta_info.h"

/* human-readable names of all of the string-type meta information values */
const char *MI_CINFO_NAMES[] = {
   "Artist",
   "Album",
   "Title",
   "Track",
   "Year",
   "Genre"
};


/* Create and return a new meta_info struct.  All memory is allocated, and
 * the resulting pointer should be free(3)'d using mi_free().
 */
meta_info *
mi_new(void)
{
   int i;

   meta_info *mi = (meta_info*) malloc(sizeof(meta_info));
   if (mi == NULL)
      err(1, "mi_new: failed to allocate meta-info");

   mi->filename = NULL;
   mi->plength_ms = 0;
   mi->last_updated = 0;
   mi->is_url = false;

   for (i = 0; i < MI_NUM_CINFO; i++)
      mi->cinfo[i] = NULL;

   return mi;
}


/* Function to free() all memory allocated by a given meta_info struct.  */
void
mi_free(meta_info *mi)
{
   int i;

   if (mi->filename != NULL)
      free(mi->filename);

   for (i = 0; i < MI_NUM_CINFO; i++) {
      if (mi->cinfo[i] != NULL)
         free(mi->cinfo[i]);
   }

   free(mi);
}

/* Function to write a meta_info struct to a file stream.  */
void
mi_fwrite(meta_info *mi, FILE *fout)
{
   static uint64_t lengths[MI_NUM_CINFO + 1];
   int i;

   /* store all necessary numeric values in the above array */
   lengths[0] = strlen(mi->filename);
   for (i = 0; i < MI_NUM_CINFO; i++) {
      if (mi->cinfo[i] == NULL)
         lengths[i+1] = 0;
      else
         lengths[i+1] = strlen(mi->cinfo[i]);
   }

   /* write */
   fwrite(lengths, sizeof(lengths), 1, fout);
   fwrite(mi->filename, sizeof(char), lengths[0], fout);
   for (i = 0; i < MI_NUM_CINFO; i++) {
      if (mi->cinfo[i] != NULL)
         fwrite(mi->cinfo[i], sizeof(char), lengths[i+1], fout);
   }
   fwrite(&(mi->plength_ms),   sizeof(uint64_t), 1, fout);
   fwrite(&(mi->last_updated), sizeof(time_t),   1, fout);
   fwrite(&(mi->is_url),       sizeof(bool),     1, fout);

   fflush(fout);
}

/* Function to read a meta_info struct from a file stream */
void
mi_fread(meta_info *mi, FILE *fin)
{
   static uint64_t lengths[MI_NUM_CINFO + 1];
   int i;

   /* first read all necessary numeric values */
   fread(lengths, sizeof(lengths), 1, fin);

   /* allocate all needed space in the meta_info struct, and zero */
   if ((mi->filename = calloc(lengths[0] + 1, sizeof(char))) == NULL)
      err(1, "mi_fread: failed to calloc filename");

   bzero(mi->filename, sizeof(char) * (lengths[0] + 1));

   for (i = 0; i < MI_NUM_CINFO; i++) {
      if (lengths[i+1] > 0) {
         if ((mi->cinfo[i] = calloc(lengths[i+1] + 1, sizeof(char))) == NULL)
            err(1, "mi_fread: failed to calloc cinfo");

         bzero(mi->cinfo[i], sizeof(char) * (lengths[i+1] + 1));
      }
   }

   /* read */
   fread(mi->filename, sizeof(char), lengths[0], fin);
   for (i = 0; i < MI_NUM_CINFO; i++) {
      if (lengths[i+1] > 0)
         fread(mi->cinfo[i], sizeof(char), lengths[i+1], fin);
   }
   fread(&(mi->plength_ms),   sizeof(uint64_t), 1, fin);
   fread(&(mi->last_updated), sizeof(time_t),   1, fin);
   fread(&(mi->is_url),       sizeof(bool),     1, fin);
}


/*****************************************************************************
 * The sanitation routines
 ****************************************************************************/
void
str_sanitize(char *s)
{
   char c;
   int  i;

   for (i = 0; i < strlen(s); i++) {
      c = s[i];
      if (!isdigit(c) && !isalpha(c) && !ispunct(c) && c != ' ')
         s[i] = '?';
   }
}

void
mi_sanitize(meta_info *mi)
{
   int i;
   for (i = 0; i < MI_NUM_CINFO; i++) {
      if (mi->cinfo[i] != NULL)
         str_sanitize(mi->cinfo[i]);
   }
}


/*****************************************************************************
 * The mi_query_* and mi_compare stuff.
 * These control setting-up/clearing the global query description, and how
 * to compare a given meta_info against this description.
 ****************************************************************************/

/* the global query description */
mi_query_description _mi_query;

/* global flag to indicate if we should match against filename in queires */
bool mi_query_match_filename;

/* initialize the query structures */
void
mi_query_init()
{
   int i;

   for (i = 0; i < MI_MAX_QUERY_TOKENS; i++)
      _mi_query.tokens[i] = NULL;

   mi_query_match_filename = true;
   _mi_query.raw = NULL;
   _mi_query.ntokens = 0;
}

/* determine if a query has been set */
bool
mi_query_isset()
{
   return _mi_query.ntokens != 0;
}

/* free the query structures */
void
mi_query_clear()
{
   int i;

   for (i = 0; i < MI_MAX_QUERY_TOKENS; i++) {
      if (_mi_query.tokens[i] != NULL) {
         free(_mi_query.tokens[i]);
         _mi_query.tokens[i] = NULL;
      }
   }

   if (_mi_query.raw != NULL) {
      free(_mi_query.raw);
      _mi_query.raw = NULL;
   }

   _mi_query.ntokens = 0;
}

/* add a token to the current query description */
void
mi_query_add_token(const char *token)
{
   if (_mi_query.ntokens == MI_MAX_QUERY_TOKENS)
      errx(1, "mi_query_add_token: reached shamefull limit");

   /* match or no? */
   if (token[0] == '!') {
      _mi_query.match[_mi_query.ntokens] = false;
      token++;
   } else
      _mi_query.match[_mi_query.ntokens] = true;

   /* copy token */
   if ((_mi_query.tokens[_mi_query.ntokens++] = strdup(token)) == NULL)
      err(1, "mi_query_add_token: strdup failed");
}

void
mi_query_setraw(const char *query)
{
   if (_mi_query.raw != NULL)
      free(_mi_query.raw);

   if ((_mi_query.raw = strdup(query)) == NULL)
      err(1, "mi_query_setraw: query strdup failed");
}

const char *
mi_query_getraw()
{
   return _mi_query.raw;
}

/* match a given meta_info struct against the global query */
bool
mi_match(const meta_info *mi)
{
   bool  matches;
   int   i, j;

   for (i = 0; i < _mi_query.ntokens; i++)
   {
      matches = false;

      /* does the filename match? */
      if (mi_query_match_filename) {
         if ((strcasestr(mi->filename, _mi_query.tokens[i])) != NULL)
            matches = true;
      }

      /* do any of the CINFO fields match? */
      for (j = 0; !matches && j < MI_NUM_CINFO; j++) {
         if (mi->cinfo[j] == NULL)
            continue;

         if ((strcasestr(mi->cinfo[j], _mi_query.tokens[i])) != NULL)
            matches = true;
      }

      if (!matches && _mi_query.match[i])
         return false;
      if (matches && !_mi_query.match[i])
         return false;
   }

   return true;
}


/*****************************************************************************
 * mi_sort_* and mi_compare stuff
 * These functions handle setting-up and clearing the global sort description
 * used when sorting meta_info objects.
 ****************************************************************************/

/* global sort description */
mi_sort_description _mi_sort;

/* initialize the sort ordering to what i like */
void
mi_sort_init()
{
   _mi_sort.order[0] = MI_CINFO_ARTIST;
   _mi_sort.order[1] = MI_CINFO_ALBUM;
   _mi_sort.order[2] = MI_CINFO_TRACK;
   _mi_sort.order[3] = MI_CINFO_TITLE;

   _mi_sort.descending[0] = false;
   _mi_sort.descending[1] = false;
   _mi_sort.descending[2] = false;
   _mi_sort.descending[3] = false;

   _mi_sort.nfields = 4;
}

/* clear the current sort */
void
mi_sort_clear()
{
   _mi_sort.nfields = 0;
}

/* Set the current sort description to what is provided in the given string.
 * The format of the string is a comma-seperated list of field-names, possibly
 * preceeded with a dash (-) to indicate that the field should be sorted
 * descending.
 * Example:
 *       s = "artist,album,track,title"
 * Would set the global sort structure to sort meta_info structs by first
 * artist, then album, then track, then title.
 * Example:
 *       s = "artist,-year"
 * Would set the global sort structure to sort meta_info struct first by
 * artist name, then by year descending (showing the most recent work of
 * a given artist first).
 *
 * RETURNS:
 *    0 if s is parsed without error.
 *    MI_ERR_SORT if s is malformed.
 */
int
mi_sort_set(const char *s)
{
   bool   found;
   char  *token;
   char  *line;
   int    idx;
   int    i;

   /* we store the new sort description here, then copy everything over.
    * if there are errors parsing, this prevents us from mucking-up
    * existing sort description.
    */
   mi_sort_description new_sort;

   if ((line = strdup(s)) == NULL)
      err(1, "mi_sort_set: sort strdup failed");

   idx = 0;
   while ((token = strsep(&line, ",")) != NULL) {

      if (strlen(token) == 0)
         continue;

      if (idx >= MI_NUM_CINFO) {
         free(line);
         return MI_ERR_SORT;
      }

      new_sort.descending[idx] = token[0] == '-';
      if (token[0] == '-') token++;

      found = false;
      for (i = 0; i < MI_NUM_CINFO && !found; i++) {
         if (strcasecmp(token, MI_CINFO_NAMES[i]) == 0) {
            new_sort.order[idx] = i;
            found = true;
         }
      }

      if (!found) {
         free(line);
         return MI_ERR_SORT;
      }

      idx++;
   }

   new_sort.nfields = idx;
   free(line);

   /* copy new sort description into global one */
   for (idx = 0; idx < new_sort.nfields; idx++) {
      _mi_sort.order[idx]      = new_sort.order[idx];
      _mi_sort.descending[idx] = new_sort.descending[idx];
   }
   _mi_sort.nfields = new_sort.nfields;

   return MI_OK;
}

/* compare two meta_info structs using the global sort description
 * Note that this function is suitable for passing to qsort(3) and the like.
 */
int
mi_compare(const void *A, const void *B)
{
   int field;
   int ret;
   int i;

   const meta_info **a2 = (const meta_info**) A;
   const meta_info **b2 = (const meta_info**) B;
   const meta_info *a = *a2;
   const meta_info *b = *b2;

   for (i = 0; i < _mi_sort.nfields; i++) {
      field = _mi_sort.order[i];

      if (a->cinfo[field] == NULL && b->cinfo[field] == NULL)
         return 0;
      if (a->cinfo[field] != NULL && b->cinfo[field] == NULL)
         return (_mi_sort.descending[i] ? 1 : -1);
      if (a->cinfo[field] == NULL && b->cinfo[field] != NULL)
         return (_mi_sort.descending[i] ? -1 : 1);

      ret = strcasecmp(a->cinfo[field], b->cinfo[field]);
      if (ret != 0)
         return (_mi_sort.descending[i] ? -1 * ret : ret);
   }

   return 0;
}


/*****************************************************************************
 * mi_display_* stuff and mi_tostr()
 ****************************************************************************/

/* global display description */
mi_display_description _mi_display;

/* initialize global display description to what i like */
void
mi_display_init()
{
   _mi_display.nfields = 5;

   _mi_display.order[0] = MI_CINFO_ARTIST;
   _mi_display.order[1] = MI_CINFO_ALBUM;
   _mi_display.order[2] = MI_CINFO_TITLE;
   _mi_display.order[3] = MI_CINFO_TRACK;
   _mi_display.order[4] = MI_CINFO_YEAR;

   _mi_display.widths[0] = 20;
   _mi_display.widths[1] = 20;
   _mi_display.widths[2] = 20;
   _mi_display.widths[3] = 4;
   _mi_display.widths[4] = 4;
}

/* reset the display to what i like */
void
mi_display_reset()
{
   mi_display_init();
}

/* return the total width of the current display description */
int
mi_display_getwidth()
{
   int   sum;
   int   i;

   sum = 0;
   for (i = 0; i < _mi_display.nfields; i++)
      sum += _mi_display.widths[i] + 1;

   return sum;
}

/* Set the global display description to the format specified in the given
 * string.
 */
int
mi_display_set(const char *display)
{
   const char *errstr = NULL;
   bool  found;
   char *s, *copy;
   char *token;
   char *num;
   int   idx;
   int   i;

   /* we build this new description here, and if everything goes
    * ok we copy it into the global display description. */
   mi_display_description new_display;
   new_display.nfields = 0;

   if ((s = strdup(display)) == NULL)
      err(1, "mi_display_set: display strdup failed");

   copy = s;
   idx = 0;

   while ((token = strsep(&s, ",")) != NULL) {

      if (strlen(token) == 0)
         continue;

      /* make sure we don't have too many fields */
      if (idx >= MI_NUM_CINFO) {
         free(copy);
         return MI_ERR_DISPLAY;
      }

      /* make sure there's a number part to the field */
      if ((num = strstr(token, ".")) == NULL) {
         free(copy);
         return MI_ERR_DISPLAY;
      }

      *num = '\0';
      num++;

      new_display.widths[idx] = strtonum(num, 1, 1000, &errstr);
      if (errstr != NULL) {
         /* TODO somehow get that errstr to rest of vitunes along with
          * whatever field we just failed to parse that for
          * Also, get rid of hard limit of 1000 above.
          */
         free(s);
         return MI_ERR_DISPLAY;
      }

      /* get the field name */
      found = false;
      for (i = 0; i < MI_NUM_CINFO && !found; i++) {
         if (strcasecmp(token, MI_CINFO_NAMES[i]) == 0) {
            new_display.order[idx] = i;
            found = true;
         }
      }

      if (!found) {
         /* TODO get a better error message to vitunes about this */
         free(copy);
         return MI_ERR_DISPLAY;
      }

      idx++;
      new_display.nfields++;
   }

   /* copy results into global display description */
   for (idx = 0; idx < new_display.nfields; idx++) {
      _mi_display.order[idx]  = new_display.order[idx];
      _mi_display.widths[idx] = new_display.widths[idx];
   }
   _mi_display.nfields = new_display.nfields;

   free(copy);
   return MI_OK;
}


/*****************************************************************************
 * begin extraction stuff...
 ****************************************************************************/

/* Simple function to help with pulling ID3 information.  libid3tag always
 * "lifts" the information returned to a ID3v2.4 format, which uses
 * unicode for textual meta info (which is all we're interested in).
 * Given an id3 tag, and the name of a "frame" (meta data atom) to pull,
 * this extracts the meta information, if there, and converts to utf8 (char).
 *
 * Parameters:
 *    tag         Pointer to id3_tag struct
 *    frame_id    String representing name of tag to extract
 *
 * Returns:
 *    IF the given frame exists in the tag, then a newly allocated string
 *    in utf8 (char) is returned.  Otherwise, or in case of any errors,
 *    NULL is returned.
 *
 * Notes:
 *    It's the responsibility of the caller to free() the string returned.
 */
char *
id3_pull_tag2char(struct id3_tag *tag, char *frame_id)
{
   struct id3_frame *frame;
   id3_ucs4_t const *strUnicode; /* unicode version of the data */
   id3_utf8_t *str;        /* good 'ol char version */

   /* check to see if frame with given ID exists in tag */
   if ((frame = id3_tag_findframe(tag, frame_id, 0)) == NULL)
      return NULL;

   /* Extract the second field (with the data) and see how many strings are
    * there.
    * NOTE: The only frames we're interested in always have two fields.
    * The first is the encoding, and the second is the actual text.
    */
   if (id3_field_getnstrings(&frame->fields[1]) < 1) {
      id3_frame_delete(frame);
      return NULL;
   }

   /* Extract just the first field */
   strUnicode = id3_field_getstrings(&frame->fields[1], 0);
   if (strUnicode == NULL) {
      id3_frame_delete(frame);
      return NULL;
   }

   /* convert to char and return ... also free stuff */
   str = id3_ucs4_utf8duplicate(strUnicode);
   if (str == NULL) {
      id3_frame_delete(frame);
      return NULL;
   }

   id3_frame_delete(frame);
   return str;
}

/*****************************************************************************
 * Each of the following functions is of the form:
 *
 *    int meta_extract_FORMAT(char *filename, meta_info *info)
 *
 * Where FORMAT is the type of meta information the function attempts to
 * extract (i.e. "id3", "mp4", "vorbis", etc.).
 *
 * For each function, the parameters and return values are interpretted the
 * same.
 *
 * Parameters:
 *    filename       String containing the file to try and extract the info
 *                   from.
 *    mi             Pointer to ALREADY ALLOCATED meta_info struct that will
 *                   be filled with any meta information found in the file.
 * Returns:
 *    IF meta information is found and extracted succesfully, 0 is returned.
 *    IF no meta information was found, but otherwise everything was
 *    successfull, then MI_ERR_NOINFO is returned.
 *    Otherwise, some error occurred while attempting to extract the info,
 *    and an appropriate error is returned.  Possible errors are:
 *
 *          MI_ERR_FOPEN      Error opening file.
 *
 *          MI_ERR_UNKNOWN    Returned when the unerlying API provides no
 *                            documentation as to why something went wrong.
 *                            Sadly, this is the most common.
 *
 ****************************************************************************/

/* Extractor for ID3 meta information */
int
meta_extract_id3(char *filename, meta_info *mi)
{
   struct id3_file  *file;
   struct id3_tag   *tag;

   /* open file and get id3 tag */
   file = id3_file_open(filename, ID3_FILE_MODE_READONLY);
   if (file == NULL)
      return MI_ERR_FOPEN;

   tag = id3_file_tag(file);
   if (tag == NULL) {
      id3_file_close(file);
      return MI_ERR_UNKNOWN;
   }

   /* pull all info */
   mi->cinfo[MI_CINFO_ARTIST] = id3_pull_tag2char(tag, ID3_FRAME_ARTIST);
   mi->cinfo[MI_CINFO_ALBUM]  = id3_pull_tag2char(tag, ID3_FRAME_ALBUM);
   mi->cinfo[MI_CINFO_TITLE]  = id3_pull_tag2char(tag, ID3_FRAME_TITLE);
   mi->cinfo[MI_CINFO_TRACK]  = id3_pull_tag2char(tag, ID3_FRAME_TRACK);
   mi->cinfo[MI_CINFO_YEAR]   = id3_pull_tag2char(tag, ID3_FRAME_YEAR);
   mi->cinfo[MI_CINFO_GENRE]  = id3_pull_tag2char(tag, ID3_FRAME_GENRE);

   /* The genre pulled above is a number (as a string) that's the index
    * into the ID3 list of genres.  here, decode that number into full
    * string.
    */
   if (mi->cinfo[MI_CINFO_GENRE] != NULL) {
      long long gindex;
      const char *errstr = NULL;

      /* note that the range is here what libid3tag can handle */
      gindex = strtonum(mi->cinfo[MI_CINFO_GENRE], 0, 147, &errstr);
      if (errstr != NULL) {  /* check for invalid genre number */
         free(mi->cinfo[MI_CINFO_GENRE]);
         mi->cinfo[MI_CINFO_GENRE] = NULL;
      } else {
         id3_ucs4_t const *gstr = id3_genre_index((unsigned int)gindex);
         free(mi->cinfo[MI_CINFO_GENRE]);
         mi->cinfo[MI_CINFO_GENRE] = id3_ucs4_utf8duplicate(gstr);
      }
   }

   /* TODO figure out how to get the total time of a track */
   mi->plength_ms = 0;

   /* cleanup */
   id3_tag_delete(tag);
   id3_file_close(file);

   /* check if we found anything */
   int i;
   for (i = 0; i < MI_NUM_CINFO; i++) {
      if (mi->cinfo[i] != NULL)
         return MI_OK;
   }

   return MI_ERR_NOINFO;
}

/* Extractor for MP4/AAC meta information */
int meta_extract_mp4(char *filename, meta_info *mi)
{
   uint16_t track1, track2;
   MP4FileHandle file;

   /* open file */
   file = MP4Read(filename, 0);
   if (file == MP4_INVALID_FILE_HANDLE)
      return MI_ERR_FOPEN;

   /* pull data.
    * NOTE: These functions will allocate each str if such a property is
    * found.  otherwise, they set the string to NULL.
    */
   MP4GetMetadataName(file,   &(mi->cinfo[MI_CINFO_TITLE]));
   MP4GetMetadataArtist(file, &(mi->cinfo[MI_CINFO_ARTIST]));
   MP4GetMetadataAlbum(file,  &(mi->cinfo[MI_CINFO_ALBUM]));
   MP4GetMetadataYear(file,   &(mi->cinfo[MI_CINFO_YEAR]));
   MP4GetMetadataTrack(file,  &track1, &track2);
   if (track1 > 0) {
      if (asprintf(&(mi->cinfo[MI_CINFO_TRACK]), "%d", track1) == -1)
         errx(1, "failure to copy mp4 track info for '%s'", filename);
   }

   /* XXX For some reason, this sometimes bombs when the genre in the m4a
    * file is specified using an ID3 style genre-id (number).
    * no clue why... been staring at the code for ages.
    * as such, no genre from mp4 files for now
    * TODO re-dig-into libmp4v2 to see why this is happening.
   MP4GetMetadataGenre(file,  &(info->genre));
    */

   /* get the duration in ms
    * NOTE: an MP4 can contain multiple "tracks", each of which as their
    * own length/time.  We take the maximum of all track lengths.
    */ 
   mi->plength_ms = 0;
   int i;
   uint32_t numTracks = MP4GetNumberOfTracks(file, NULL, 0);
   for (i = 0; i < numTracks; i++) {
      MP4TrackId track      = MP4FindTrackId(file, i, NULL, 0);
      MP4Duration numTicks  = MP4GetTrackDuration(file, track);
      uint64_t time = MP4ConvertFromTrackDuration(file, track, numTicks,
         MP4_USECS_TIME_SCALE);

      if (time > mi->plength_ms)
         mi->plength_ms = time;
   }

   /* cleanup */
   MP4Close(file);

   /* check if we found anything */
   for (i = 0; i < MI_NUM_CINFO; i++) {
      if (mi->cinfo[i] != NULL)
         return MI_OK;
   }

   return MI_ERR_NOINFO;
}

/* Extractor for OGG/VORBIS meta information */
int meta_extract_vorbis(char *filename, meta_info *mi)
{
   OggVorbis_File file;
   vorbis_comment *vcomments;

   /* open file and pull comments */
   if (ov_fopen(filename, &file) < 0)
      return MI_ERR_FOPEN;

   vcomments = ov_comment(&file, -1);
   if (vcomments == NULL) {
      ov_clear(&file);
      return MI_ERR_UNKNOWN;
   }

   /* pull info */
   int i;
   for (i = 0; i < vcomments->comments; i++) {
      char *value = strchr(vcomments->user_comments[i], '=') + 1;

      if (strncasecmp(vcomments->user_comments[i], "artist", 6) == 0)
         asprintf(&(mi->cinfo[MI_CINFO_ARTIST]), "%s", value);
      else if (strncasecmp(vcomments->user_comments[i], "album", 5) == 0)
         asprintf(&(mi->cinfo[MI_CINFO_ALBUM]), "%s", value);
      else if (strncasecmp(vcomments->user_comments[i], "title", 5) == 0)
         asprintf(&(mi->cinfo[MI_CINFO_TITLE]), "%s", value);
      else if (strncasecmp(vcomments->user_comments[i], "tracknumber", 11) == 0)
         asprintf(&(mi->cinfo[MI_CINFO_TRACK]), "%s", value);
      else if (strncasecmp(vcomments->user_comments[i], "date", 4) == 0)
         asprintf(&(mi->cinfo[MI_CINFO_YEAR]), "%s", value);
      else if (strncasecmp(vcomments->user_comments[i], "genre", 5) == 0)
         asprintf(&(mi->cinfo[MI_CINFO_GENRE]), "%s", value);
   }

   /* get the time */
   mi->plength_ms = 0;
   double time = ov_time_total(&file, -1);
   if (time != OV_EINVAL)
      mi->plength_ms = (uint64_t)(time * 1000000);

   /* cleanup */
   ov_clear(&file);
   /* XXX does the vorbis_comment struct need to be free()'d?  Or is it
    * setup to point to existing structs in the oggfile struct?
    * Documentation is no help, and I see not ov_comment_free() function
    * in the library.
    */

   /* check if we found anything */
   for (i = 0; i < MI_NUM_CINFO; i++) {
      if (mi->cinfo[i] != NULL)
         return MI_OK;
   }

   return MI_ERR_NOINFO;
}

/* The main top-level function of this module.  It's also the only function
 * (aside from meta_info_new and meta_info_free) used outside of this module.
 * Given a the name of a file, this function will attempt to extract any
 * meta information from the file and return a pointer to a newly allocated
 * meta_info struct.  If unsuccessfull, NULL is returned.
 *
 * Parameters:
 *    filename    Name of file to extract info from.
 *
 * Returns:
 *    If successfull in extracting meta information, a pointer to a newly
 *    allocated meta_info struct is returned.
 *    If unsuccessfull, for ANY reason, NULL is returned.
 *
 * Notes:
 *    It's the responsibility of the caller to free the meta_info struct
 *    returned using the meta_info_free function.
 *
 *    The filename given will be expanded using realpath(3).  If that fails,
 *    a warning is displayed to stderr and NULL is returned, but processing
 *    will continue.
 *
 *    If a malloc() within this function fails, the program is halted
 *    immediately.
 *
 *    Any other error (namely within the meta_extract_FORMAT functions) are
 *    simply ignored, and there is currently no way to querry what/where the
 *    error was. TODO
 */
meta_info *
meta_extract(const char *filename)
{
   static char fullname[PATH_MAX];

   /* create meta info struct */
   meta_info *mi = mi_new();

   /* store full filename in meta_info struct */
   bzero(fullname, sizeof(fullname));
   if (realpath(filename, fullname) == NULL)
      err(1, "couldn't resolve '%s' to full path name", filename);

   if (asprintf(&(mi->filename), "%s", fullname) == -1)
      errx(1, "failure to copy filename '%s' for meta info", fullname);

   time(&mi->last_updated);

   /* check each of the formats supported in an arbitrary order.
    * if/when any of them find meta info, return immediately.
    * NOTE: if any of the extraction function encounters any error, we
    * simply keep trying every other function.
    */

   if (meta_extract_id3(fullname, mi) == 0)
      return mi;

   if (meta_extract_vorbis(fullname, mi) == 0)
      return mi;

   if (meta_extract_mp4(fullname, mi) == 0)
      return mi;

   /* if we've reached here, no meta info found.  free the meta_info
    * struct and return null.
    */
   mi_free(mi);
   return NULL;
}
