/*
 * Copyright (c) 2010 Ryan Flannery <ryan.flannery@gmail.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "player.h"

/* global struct used to monitor the player */
mp_status_t mp_status;

/* global media player */
media_player player;


/*****************************************************************************
 * player setup/destroy functions
 ****************************************************************************/

/* Initialize the media player struct.  This includes fork()'ing the mplayer
 * child process and setting up read/write ends of pipe to it.
 */
void
player_init()
{
   int   pwrite[2];
   int   pread[2];
   pid_t child_pid;

   /* ignore broken pipe signal -- event is handled elsewhere */
   signal(SIGPIPE, SIG_IGN);

   /* create the pipes */
   if (pipe(pwrite) == -1 || pipe(pread) == -1) {
      perror("failure creating pipes for media player");
      exit(1);
   }

   /* fork the child */
   child_pid = fork();
   switch (child_pid) {
      case -1:
         err(1, "couldn't fork() media player");
         break;

      case 0:  /* child process */
         if (close(0) == -1 || close(1) == -1 || close(2) == -1)
            err(1, "failed in fork()'ed child to close stdin/stdout");

         if (close(pwrite[1]) == -1 || close(pread[0]) == -1)
            err(1, "failed in fork()'ed child to close excess pipes");

         if (dup(pwrite[0]) == -1 || dup(pread[1]) == -1)
            err(1, "failed in fork()'ed child dup()'ing pipe");

         /* launch the media player */
         /* TODO: fuck mplayer -- try to find a better msglevel */
         /* TODO: update - with '-ao sndio' the output is much, much less
          *       frequent.
          */
         if (execl("/usr/local/bin/mplayer", "mplayer",
               "-slave", "-idle", "-quiet", "-msglevel",
               "cplayer=0:ao=0:vo=0:decaudio=0:decvideo=0:demuxer=0",
               (char*)NULL) == -1)
            err(1, "failed to exec the media player");

         /* NOT REACH in fork()'d child */
         break;
   }

   /* back to the parent... */
   if (close(pwrite[0]) == -1 || close(pread[1]) == -1)
      err(1, "failed to close excess pipes");

   /* set player */
   player.pid        = child_pid;
   player.pipe_read  = pread[0];
   player.pipe_write = pwrite[1];

   player.queue      = NULL;
   player.current    = -1;

   player.mode       = MP_MODE_LOOP;
   player.rand_seed  = time(0);
   srand(player.rand_seed);

   player.should_be_playing    = false;
   player.paused               = false;
   player.just_started_playing = false;

   /* setup player status object */
   mp_status.is_playing = false;
   mp_status.position = 0;

   /* setup read pipe to media player as non-blocking */
   int flags = fcntl(player.pipe_read, F_GETFL, 0);
   if (flags == -1)
      err(1, "failed to fcntl() the read-pipe to media player (get)");

   if (fcntl(player.pipe_read, F_SETFL, flags | O_NONBLOCK) == -1)
      err(1, "failed to fcntl() the read-pipe to media player (set)");
}

/* Kill the child process that is mplayer */
void
player_shutdown()
{
   static const char *cmd = "\nquit\n";

   /* reset status object */
   mp_status.is_playing = false;
   mp_status.position = 0;

   /* instruct media player to exit */
   if (write(player.pipe_write, cmd, strlen(cmd)) == -1) {
      if (errno == EPIPE)
         return;  /* child is already dead - nothing more to do */
      else
         err(1, "failed to send quit command to media player");
   }

   /* close pipes and wait for media player to die */
   close(player.pipe_read);
   close(player.pipe_write);
   waitpid(player.pid, NULL, 0);
   player.queue = NULL;
   player.current = -1;
}


/*****************************************************************************
 * player management functions (setting queue, playing, pausing, etc.)
 ****************************************************************************/

/* setup the given playlist to be played by the media player */
void
player_set_queue(playlist *queue)
{
   player_stop();

   player.queue   = queue;
   player.current = -1;
   if (player.queue->nfiles > 0)
      player.current = 0;
}

/* start playing */
void
player_play()
{
   static char response[1000];
   char       *cmd;

   if (player.queue == NULL)
      errx(1, "media queue is null");

   if (player.current < 0 || player.current > player.queue->nfiles)
      errx(1, "media item to play is out of range!");

   /* flush */
   read(player.pipe_read, &response, 1000);
   read(player.pipe_read, &response, 1000);

   /* build the command to send to mplayer and send it */
   asprintf(&cmd, "\nloadfile \"%s\" 0\nget_time_pos\n",
      player.queue->files[player.current]->filename);
   if (cmd == NULL)
      err(1, "player_play() failed to build command");

   write(player.pipe_write, cmd, strlen(cmd));
   free(cmd);

   /* update player flags */
   player.just_started_playing = true;
   player.should_be_playing = true;
   player.paused            = false;

   /* update status */
   mp_status.is_playing = true;
   mp_status.position = 0;
}

/* cease playback of mplayer */
void
player_stop()
{
   /* XXX FUCK MPLAYER -- yah, this is documented and seems reasonable, but
    * has no affect what-so-ever ...
   static const char *cmd = "\nstop\n";
    * so instead i use the following hack to stop mplayer... load a file
    * that does not exist (well, hopefully there's no file named
    * 'dummy.dummy.dummy').
    */
   static const char *cmd = "\nloadfile dummy.dummy.dummy\n";

   write(player.pipe_write, cmd, strlen(cmd));

   /* flush */
   static char response[1000];
   read(player.pipe_read, &response, 1000);
   read(player.pipe_read, &response, 1000);

   /* update status flags */
   player.just_started_playing = false;
   player.should_be_playing = false;
   player.paused = false;
   mp_status.is_playing = false;
}

/* pause/unpause playback of mplayer */
void
player_pause()
{
   static const char *cmd = "\npause\n";

   write(player.pipe_write, cmd, strlen(cmd));

   player.paused = !player.paused;
}

/* seek forward/backward in the currently playing file */
void
player_seek(int seconds)
{
   char *cmd;

   if (!player.should_be_playing || player.paused)
      return;

   asprintf(&cmd, "\nseek %i 0\nget_time_pos\n", seconds);
   if (cmd == NULL)
      err(1, "player_seek() failed to build command");

   write(player.pipe_write, cmd, strlen(cmd));
   free(cmd);
}

/*****************************************************************************
 * player monitor function
 ****************************************************************************/

/* communicate with the child mplayer process and update global mp_status
 * object with information about if mplayer is player, if so, how far it is
 * through the file.
 */
void
player_monitor(void)
{
   static const char *info_cmds = "\nget_time_pos\n";
   static const int response_size = 1000;
   static char  response[1000];

   char *response_start;
   int bytes_read;

   /* fuck mplayer, fuck it to hell */
   if (player.paused) return;

   /* query the media player for info */
   bzero(response, response_size * sizeof(char));
   bytes_read = read(player.pipe_read, &response, response_size);

   if (bytes_read > 0) {
      /* sanitize the response and sift through any crap */
      /* TODO there may be multiple ANS_TIME_POSITION responses in the
       * buffer read ... figure how to get to the last one... */
      response[bytes_read + 1] = 0;
      response_start = strstr(response, "ANS_TIME_POSITION");

      /* if an answer to the position query was found, strip it */
      if (response_start != NULL) {
         if (sscanf(response_start, "ANS_TIME_POSITION=%f",
                &mp_status.position) != 1)
                  errx(1, "mplayer is misbehaving");

         player.should_be_playing = true;
         player.paused            = false;
         player.just_started_playing = false;
         mp_status.is_playing = true;
      }

   } else if (bytes_read == -1) {

      if (errno == EAGAIN && player.should_be_playing && !player.paused) {

         if (player.just_started_playing)
            player.just_started_playing = true;
         else if ((player.current = get_next_song_index()) == -1) {
            /* reached end of playlist */
            player.should_be_playing = false;
            player.paused = false;
            mp_status.is_playing = false;
         } else
            player_play();
      }
   }

   /* Re-query the media player for more stats. */
   if (!player.paused)
      write(player.pipe_write, info_cmds, strlen(info_cmds));
}

int
get_next_song_index()
{
   if (player.queue == NULL)
      errx(1, "get_next_song_index() empty queue!");

   switch (player.mode) {
      case MP_MODE_LINEAR:
         if (player.current == player.queue->nfiles - 1)
            return -1;
         else
            return player.current + 1;
         break;

      case MP_MODE_LOOP:
         if (player.current == player.queue->nfiles - 1)
            return 0;
         else
            return player.current + 1;
         break;

      case MP_MODE_RANDOM:
         return rand_r(&(player.rand_seed)) % player.queue->nfiles;
         break;
   }
   return 0;
}

