#ifndef PLAYLIST_H
#define PLAYLIST_H

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glob.h>
#include <unistd.h>
#include <libgen.h>
#include <errno.h>
#include <err.h>

#include "meta_info.h"
#include "debug.h"

#define PLAYLIST_CHUNK_SIZE   100

/* the core playlist structure */
typedef struct
{
   char  *filename;  /* filename containing the playlist */
   char  *name;      /* name of the playlist used in display */

   /* the array of files (their meta information) in the playlist */
   meta_info **files;
   int         nfiles;     /* number of files in the playlist */
   int         capacity;   /* current size malloc()'d for the files */

   /* This is a flag for playlists indicating if they have any unsaved
    * changes.  Since psuedo playlists are excluded, this flag is
    * managed entirely by input_handlers.c, except for being set to a
    * default of false in playlist_new.
    * It is also used in paint.c.
    */
   bool        needs_saving;

} playlist;

/* IMPORTANT NOTES ABOUT THE "playlist" STRUCTURE:
 * 1. The elements of the "files" array are simply pointers to the
 *    already existing meta-info elements in the media database.
 *
 * 2. When loading a playlist from a file, each element of the playlist
 *    is compared against the media database to find a corresponding entry.
 *    If no such file exists in the media database, a new record is added
 *    to the DB but it contains *only* the filename read from the playlist
 *    file (no meta info).
 *
 * 3. The media database mentioned above is simply an array of meta_info
 *    structs, passed to the functions when necessary below.
 */

/* create/destroy/duplicate playlist structs */
playlist *playlist_new(void);
void playlist_free(playlist *p);
playlist *playlist_dup(const playlist *original, const char *filename,
                       const char* name);

/* remove a file from a playlist */
void playlist_file_add(playlist *p, meta_info *f, int index);
void playlist_file_append(playlist *p, meta_info *f);
void playlist_file_remove(playlist *p, int index);
void playlist_file_replace(playlist *p, int index, meta_info *newEntry);

/* load/save/delete playlists from/to/from filesystem */
playlist *playlist_load(const char *filename, meta_info **db, int ndb);
void playlist_save(const playlist *p);
void playlist_delete(playlist *p);

/* filter a playlist to all records matching/not-matching a given string */
playlist *playlist_filter(const playlist *p, bool m);

/* retrieve all playlist files in a given directory and return number found */
int retrieve_playlist_filenames(const char *dirname, char ***files);

#endif
