#include "vitunes.h"
#include "config.h"

/* fixed locations of various things (the %s in all is the user's homedir) */
const char *VITUNES_DIR_FMT  = "%s/.vitunes";
const char *CONF_FILE_FMT    = "%s/.vitunes/vitunes.conf";
const char *DB_FILE_FMT      = "%s/.vitunes/vitunes.db";
const char *PLAYLIST_DIR_FMT = "%s/.vitunes/playlists";

char *vitunes_dir;
char *conf_file;
char *db_file;
char *playlist_dir;

/* used with -DDEBUG */
FILE  *debug_log;

/* global var to track when/if to quit vitunes */
bool  QUIT_VITUNES;

/* global pointers to currently viewed and playing playlists */
playlist *viewing_playlist;
playlist *playing_playlist;

/* misc. funcitons */
void usage(const char *);
void sig_winch(int);
void sig_timer(int);
void setup_timer();
void load_config();
int  handle_switches(int argc, char *argv[]);

/* e-command handlers */
int  ecmd_init(int, char *a[]);
int  ecmd_add(int, char *a[]);
int  ecmd_addurl(int, char *a[]);
int  ecmd_check(int, char *a[]);
int  ecmd_rmfile(int, char *a[]);
int  ecmd_update(int, char *a[]);
int  ecmd_flush(int, char *a[]);
int  ecmd_help(int, char *a[]);
void ecmd_help_init();
void ecmd_help_add();
void ecmd_help_addurl();
void ecmd_help_check();
void ecmd_help_rmfile();
void ecmd_help_update();
void ecmd_help_flush();

/* build e-command path */
struct ecmd_entry {
   char  *name;
   int   (*ecmd)(int argc, char *argv[]);
   void  (*help)(void);
};

const struct ecmd_entry ECMD_PATH[] = {
   { "init",      ecmd_init,     ecmd_help_init},
   { "add",       ecmd_add,      ecmd_help_add },
   { "addurl",    ecmd_addurl,   ecmd_help_addurl },
   { "check",     ecmd_check,    ecmd_help_check },
   { "rmfile",    ecmd_rmfile,   ecmd_help_rmfile },
   { "update",    ecmd_update,   ecmd_help_update },
   { "flush",     ecmd_flush,    ecmd_help_flush },
   { "help",      ecmd_help,     NULL }
};
const int ECMD_PATH_SIZE = (sizeof(ECMD_PATH) / sizeof(struct ecmd_entry));



int
main(int argc, char *argv[])
{
   struct passwd *pwd;
   int    input;
   int    previous_command;
   int    i;

#ifdef DEBUG
   if ((debug_log = fopen("log.txt", "w")) == NULL)
      err(1, "failed to open debug log");
#endif

   /*------------------------------------------------------------------------
    * build directory/file names needed by vitunes & handle arguments
    *----------------------------------------------------------------------*/

   /* get home dir and build locations of various files */
   pwd = getpwuid(getuid());
   if (pwd == NULL)
      errx(1, "invalid user %d", getuid());

   /* build paths */
   asprintf(&vitunes_dir,   VITUNES_DIR_FMT,  pwd->pw_dir);
   asprintf(&conf_file,     CONF_FILE_FMT,    pwd->pw_dir);
   asprintf(&db_file,       DB_FILE_FMT,      pwd->pw_dir);
   asprintf(&playlist_dir,  PLAYLIST_DIR_FMT, pwd->pw_dir);
   if (vitunes_dir == NULL || conf_file == NULL
   ||  db_file == NULL     || playlist_dir == NULL)
      err(1, "failed to create needed file names");

   /* handle command-line switches ... also handles logic for running all of
    * the e-commands */
   handle_switches(argc, argv);


   /*------------------------------------------------------------------------
    * initialize stuff
    *--------------------------------------------------------------------- */

   mi_query_init();        /* global query description */
   mi_sort_init();         /* global sort description */
   mi_display_init();      /* global display description */
   ybuffer_init();         /* global yank/copy buffer */
   search_dir_init();      /* global search direction */
   gnum_clear();           /* global input number */

   /* load media library (database and all playlists) */
   medialib_load(db_file, playlist_dir);

   /* check here to see if database is empty...if so, don't load vitunes */
   if (mdb.library->nfiles == 0) {
      warnx("the vitunes database is currently empty.  see 'vitunes -e help add' for how to add files.");
      return 0;
   }

   /* setup user interface and default colors */
   ui_init(DEFAULT_LIBRARY_WINDOW_WIDTH);
   paint_setup_colors();

   /* basic ui setup to get ui started */
   playing_playlist = NULL;
   ui.library->nrows  = mdb.nplaylists;
   setup_viewing_playlist(mdb.library);
   ui_draw_borders();
   qsort(mdb.library->files, mdb.library->nfiles,
      sizeof(meta_info*), mi_compare);


   /* load config file and run commnads in it now */
   load_config();

   /* now start mplayer */
   player_init();

   /* initial painting of the display */
   paint_all();

   /* setup signal handlers */
   signal(SIGWINCH, sig_winch);  /* resize */
   setup_timer();                /* periodic timer to update mplayer */

   /* FIXME: so this successfully ignores the control+y delayed suspend
    * signal, but i'm still unable, apparently, to capture this in curses.
    * at least "getch()" appears to return nothing in response to control+y
    * even with this signal being ignored.  I could do something more
    * complicated (set a global flag)... but I'd like to investigate this
    * further first...
   */
   signal(SIGTSTP, SIG_IGN);

   /* -----------------------------------------------------------------------
    * begin input loop
    * -------------------------------------------------------------------- */

   QUIT_VITUNES = false;
   previous_command = -1;

   while ((input = getch()) && !QUIT_VITUNES)
   {
      /* no input, just skip */
      if (input == ERR)
         continue;

      /* handle input */
      if (isdigit(input) &&  (input != '0' || gnum_get() > 0))
         gnum_add(input - '0');
      else if (input == '\n' && gnum_get() > 0 && previous_command >= 0)
         ExecuteKeyBinding(previous_command);
      else {
         for (i = 0; i < KeyBindingsSize; i++) {
            if (KeyBindings[i].keycode == input) {
               ExecuteKeyBinding(i);
               previous_command = i;
            }
         }
      }

   }

   /* -----------------------------------------------------------------------
    * clenaup
    * -------------------------------------------------------------------- */

   /* ignore all registered signals */
   signal(SIGALRM,  SIG_IGN);
   signal(SIGWINCH, SIG_IGN);

   /* cleanup */
   ui_destroy();
   medialib_destroy();
   player_shutdown();

   mi_query_clear();
   ybuffer_free();

   return 0;
}

/* print proper usage */
void
usage(const char *pname)
{
   fprintf(stderr,"\
usage: %s [-f config-file] [-d database-file] [-p playlist-dir] [-e COMMAND ...]\n\
See \"%s -e help\" for information about what e-commands are available.\n\
",
   pname, pname);
   exit(1);
}

/* handle re-size signals */
void
sig_winch(int dummy)
{
   ui_resize();
   ui_clear();
   paint_all();
}

/* signal handler for timer.
 * updates mplayer status and repaints windows if needed
 */
void
sig_timer(int dummy)
{
   static playlist *prev_queue = NULL;
   static int       prev_index = 0;

   /* update what mplayer is doing */
   player_monitor();

   /* always paint player */
   paint_player();

   /* need to re-paint library window? */
   if (player.queue != prev_queue) {
      prev_queue = player.queue;
      paint_library();
   }

   /* need to re-paint playlist window? */
   if (player.queue == viewing_playlist && prev_index != player.current) {
      prev_index = player.current;
      paint_playlist();
   }
}

/* setup timer signal handler above */
void
setup_timer()
{
   struct sigaction sig_act;
   struct itimerval timer;

   /* create timer signal handler */
   if (sigemptyset(&sig_act.sa_mask) < 0)
      err(1, "setup_timer: sigemptyset failed");

   sig_act.sa_flags = 0;
   sig_act.sa_handler = sig_timer;
   if (sigaction(SIGALRM, &sig_act, NULL) < 0)
      err(1, "setup_timer: sigaction failed");

   /* setup timer details */
   timer.it_value.tv_sec = 0;
   timer.it_value.tv_usec = 500000;
   timer.it_interval.tv_sec = 0;
   timer.it_interval.tv_usec = 500000;

   /* register timer */
   if (setitimer(ITIMER_REAL, &timer, NULL) < 0)
      err(1, "setup_timer: setitimer failed");
}

/*****************************************************************************
 *
 *                            E - Commands
 *
 * What follows are all of the "E-Command" functions... that is, the functions
 * that are executed when "vitunes -e commandname" is entered.
 * The name of each function is preceeded with "ecmd_", and each takes the
 * form of a argc+argv[] argument list.
 * Note that their return values are the return values for this program.
 *
 ****************************************************************************/

int
ecmd_init(int argc, char *argv[])
{
   if (argc != 1)
      errx(1, "usage: -e %s", argv[0]);

   printf("Creating all necessary files and directories for vitunes...\n");
   medialib_setup_files(vitunes_dir, db_file, playlist_dir);

   printf("\nNow use 'vitunes -e add dir1 dir2 ...' to add files to vitunes\n");
   return 0;
}

int
ecmd_update(int argc, char *argv[])
{
   if (argc != 1)
      errx(1, "usage: -e %s", argv[0]);

   printf("Loading existing database...\n");
   medialib_load(db_file, playlist_dir);

   printf("Updating existing database...\n");
   medialib_db_update();

   medialib_destroy();
   return 0;
}

int
ecmd_add(int argc, char *argv[])
{
   if (argc == 1)
      errx(1, "usage: -e %s /path/to/filesORdirs [ ... ] ", argv[0]);

   printf("Loading existing database...\n");
   medialib_load(db_file, playlist_dir);

   printf("Scanning directories for files to add to database...\n");
   medialib_db_scan_dirs(argv + 1);

   medialib_destroy();
   return 0;
}

int
ecmd_addurl(int argc, char *argv[])
{
   meta_info   *m;
   bool         found;
   int          found_idx;
   char         input[255];
   int          field, i;

   if (argc != 2)
      errx(1, "usage: -e %s filename|URL", argv[0]);

   /* start new record, set filename */
   m = mi_new();
   m->is_url = true;
   if ((m->filename = strdup(argv[1])) == NULL)
      err(1, "ecmd_addurl: strdup failed (filename)");

   /* get fields from user */
   for (field = 0; field < MI_NUM_CINFO; field++) {

      printf("%10.10s: ", MI_CINFO_NAMES[field]);
      if (fgets(input, sizeof(input), stdin) == NULL) {
         warnx("operation cancelled.  database unchanged.");
         mi_free(m);
         return 0;
      }

      if (input[strlen(input) - 1] == '\n')
         input[strlen(input) - 1] = '\0';

      if ((m->cinfo[field] = strdup(input)) == NULL)
         err(1, "ecmd_addurl: strdup failed (field)");
   }

   /* load existing databae and see if file/url already exists */
   medialib_load(db_file, playlist_dir);

   /* does the url already exist in the database? */
   found = false;
   found_idx = -1;
   for (i = 0; i < mdb.library->nfiles && !found; i++) {
      if (strcmp(m->filename, mdb.library->files[i]->filename) == 0) {
         found = true;
         found_idx = i;
      }
   }

   if (found) {
      printf("warning: file/url '%s' already in the database.\n", argv[0]);
      printf("Do you want to replace the existing record? [y/n] ");

      if (fgets(input, sizeof(input), stdin) == NULL
      || (strcasecmp(input, "yes\n") != 0 && strcasecmp(input, "y\n") != 0)) {
         warnx("operation cancelled.  database unchanged.");
         mi_free(m);
         medialib_destroy();
         return 0;
      }

      mi_sanitize(m);
      playlist_file_replace(mdb.library, found_idx, m);
   } else {
      mi_sanitize(m);
      playlist_file_append(mdb.library, m);
   }

   medialib_db_save(db_file);
   medialib_destroy();

   return 0;
}

int
ecmd_check(int argc, char *argv[])
{
   meta_info *mi;
   bool found;
   int  i;

   if (argc != 2)
      errx(1, "usage: -e %s check filename", argv[0]);

   /* attempt to extract info from file */
   mi = meta_extract(argv[1]);

   /* did we get any info? */
   if (mi == NULL)
      warnx("failed to extract any information from '%s'", argv[1]);
   else {
      /* got info... show raw data, then sanitized */

      printf("The following RAW meta information was extracted:\n");
      for (i = 0; i < MI_NUM_CINFO; i++)
         printf("%10.10s: '%s'\n", MI_CINFO_NAMES[i], mi->cinfo[i]);

      mi_sanitize(mi);

      printf("The following SANITIZED meta information was extracted:\n");
      for (i = 0; i < MI_NUM_CINFO; i++)
         printf("%10.10s: '%s'\n", MI_CINFO_NAMES[i], mi->cinfo[i]);
   }

   /* check if it's in the database */
   printf("Checking if it exists in the database...\n");
   medialib_load(db_file, playlist_dir);
   found = false;
   for (i = 0; i < mdb.library->nfiles; i++) {
      if (strcmp(argv[1], mdb.library->files[i]->filename) == 0)
         found = true;
   }

   if (found)
      printf("File exists in database.\n");
   else
      printf("File does NOT exist in database.\n");

   medialib_destroy();

   return 0;
}

int
ecmd_rmfile(int argc, char *argv[])
{
   char *filename;
   char  input[255];
   bool  forced;
   bool  found;
   int   found_idx;
   int   i;

   if (argc < 2 || argc > 3)
      errx(1, "usage: -e %s [-f] filename|URL", argv[0]);

   /* is this remove forced? */
   forced = false;
   if (argc == 3) {
      if (strcmp(argv[1], "-f") != 0)
         errx(1, "usage: -e %s [-f] filename|URL", argv[0]);
      else
         forced = true;

      filename = argv[2];
   } else
      filename = argv[1];


   medialib_load(db_file, playlist_dir);

   /* search for the record */
   found = false;
   found_idx = -1;
   for (i = 0; i < mdb.library->nfiles && !found; i++) {
      if (strcmp(filename, mdb.library->files[i]->filename) == 0) {
         found = true;
         found_idx = i;
      }
   }

   if (!found) {
      i = (forced ? 0 : 1);
      errx(i, "%s: %s: No such file or url", argv[0], filename);
   }

   if (!forced) {
      printf("Are you sure you want to delete '%s'? [y/n] ", filename);
      if (fgets(input, sizeof(input), stdin) == NULL
      || (strcasecmp(input, "yes\n") != 0 && strcasecmp(input, "y\n") != 0))
         errx(1, "operation cancelled.  database unchanged.");
   }

   playlist_file_remove(mdb.library, found_idx);
   medialib_db_save(db_file);
   medialib_destroy();
   return 0;
}

int
ecmd_flush(int argc, char *argv[])
{
   if (argc != 1)
      errx(1, "usage: -e %s", argv[0]);

   medialib_load(db_file, playlist_dir);
   /* TODO medialib_db_flush(stdout);*/
   printf("TODO...er, TO-re-DO\n");
   medialib_destroy();
   return 0;
}

void
ecmd_help_init(void)
{  printf("\
VITUNES COMMAND:\n\tinit - initialize vitunes directories/databaes\n\n\
SYNOPSIS:\n\tinit\n\n\
DESCRIPTION:\n\
   The init command is used to setup the initial files and directories\n\
   used by vitunes.  This includes:\n\n\
      ~/.vitunes              The vitunes core directory where everything\n\
                              is stored.\n\
      ~/.vitunes/vitunes.db   The library database containing the meta\n\
                              information of all media files.\n\
      ~/.vitunes/playlists/   Directory where all playlists are stored.\n\n\
   The database created is initially empty.\n\
   If any of the above files/directories exists, nothing will be changed.\n\
   This command takes no parameters.\n\n\
EXAMPLE:\n\
      $ vitunes -e init\n\n\
");
}

void
ecmd_help_add(void)
{  printf("\
VITUNES COMMAND:\n\tadd - add files to the vitunes database\n\n\
SYNOPSIS:\n\tadd /path/to/files1 [ path2 ... ]\n\n\
DESCRIPTION:\n\
   The add command is used to add files to the database used by vitunes.\n\
   For every file/directory provided as a parameter, vitunes will scan that\n\
   file or directory (recursively) searching for media files that contain\n\
   any meta information.  Any such files found are added to the database\n\
   used by vitunes.\n\n\
   Note that vitunes only maintains information about the file, and not the\n\
   file itself.  It does NOT move/copy/modify the files in its database in\n\
   any way.\n\n\
   The information vitunes stores for each file includes:\n\
      *  Filename                *  Track Number\n\
      *  Artist Name             *  Year\n\
      *  Album Name              *  Genre\n\
      *  Song/Video Title\n\n\
   The filename stored is the absolute pathname obtained from realpath(3),\n\
   and serves as the key-field within the database.\n\n\
   If any file encountered has no meta information, it is NOT added to the\n\
   database.\n\n\
EXAMPLE:\n\
   $ vitunes -e add ~/music /usr/local/share/music\n\n\
");
}

void
ecmd_help_addurl(void)
{
   printf("\
VITUNES COMMAND:\n\taddurl - add a URL (or other) to the vitunes database\n\n\
SYNOPSIS:\n\taddurl URL\n\n\
DESCRIPTION:\n\
   To add non-standard-files to the vitunes database (things like URL's for\n\
   internet radio streams), one can use the addurl command. It takes a single\n\
   parameter: the url/filename to be added to the database. After that, you\n\
   will be prompted to enter meta-information for each field vitunes indexes\n\
   (artist, album, title, track, year, and genre). You may leave any/all of\n\
   the fields blank.\n\n\
   Basically, anything \"foo\" that mplayer can play by a simple:\n\
      $ mplayer foo\n\
   can be added to the database using this command.  Although regular files\n\
   could also be added using this command, the add commnad is prefered, as\n\
   it attempts to extract meta information automatically.\n\n\
   Note that files added to the database using the addurl commnad are NOT\n\
   checked for updates during an update command.\n\n\
EXAMPLE:\n\
   $ vitunes -e addurl \"http://198.234.121.118/listen.pls\"\n\
   Artist: WVXU Online Radio<ENTER>\n\
    Album: Cincinnati Public Radio<ENTER>\n\
    Title: NPR<ENTER>\n\
    Track: <ENTER>\n\
     Year: <ENTER>\n\
    Genre: Radio<ENTER>\n\n\
");
}

void
ecmd_help_check(void)
{
   printf("\
VITUNES COMMAND:\n\tcheck - check a single file for meta info and if it's in db\n\n\
SYNOPSIS:\n\tcheck filename\n\n\
DESCRIPTION:\n\
   To check an individual file to see if vitunes can extract any meta-\n\
   information, or to see how vitunes \"sanitizes\" the meta information\n\
   (see below), one can use the check command.\n\n\
   vitunes will attempt to extract any meta information from the file and\n\
   then write to stdout:\n\
      1) the raw meta information it was able to extract,\n\
      2) the sanitized meta information, and\n\
      3) whether or not the file exists in the vitunes database.\n\n\
SANITATION:\n\
   By \"sanitiation\" above, the following is meant... some media files have\n\
   non-printable characters in their meta information for one reason or\n\
   another (often, it's a faulty ripper/tagging application).  Sometimes\n\
   these non-printable characters can be control sequences used by [n]curses\n\
   and thus cause problems with the curses display of vitunes.  The way\n\
   vitunes sanitizes such data is to replace all such control characters with\n\
   an '?'.\n\n\
EXAMPLE:\n\
   $ vitunes -e check /path/to/file.mp3\n\n\
");
}

void
ecmd_help_rmfile(void)
{
   printf("\
VITUNES COMMAND:\n\trmfile - remove a file/URL from the vitunes database\n\n\
SYNOPSIS:\n\trmfile [-f] filename/URL\n\n\
DESCRIPTION:\n\
   To remove a file/URL from the vitunes database, the rmfile command may\n\
   be used.  It takes a single parameter: the filename/URL of the file to\n\
   remove.  Normally, you will be prompted if you are sure you want to\n\
   remove the file as a safety measure.  This prompt can be avoided using\n\
   the '-f' flag, to force the removal.\n\n\
   Note that to remove files, the full, absolute path to the file must be\n\
   provided, as obtained from realpath(3).\n\n\
EXAMPLE:\n\
   $ vitunes -e rmfile -f \"http://198.234.121.118/listen.pls\"\n\n\
");
}

void
ecmd_help_update(void)
{  printf("\
VITUNES COMMAND:\n\tupdate - update vitunes database\n\n\
SYNOPSIS:\n\tupdate\n\n\
DESCRIPTION:\n\
   The update command loads the existing meta information database used\n\
   by vitunes and for each media file listed in that database, the file is\n\
   checked to see if it has been removed or modified since it was added\n\
   to the database.\n\n\
   If the file has been removed, it will be removed from the database.\n\
   If the file has been modified, it's meta information will be extracted\n\
   again and the database will be updated.\n\
   If there are errors checking the file, the error will be reported but\n\
   the file, and its meta information, will remain in the vitunes database.\n\n\
   In short, anytime you remove/modify media files already in the vitunes\n\
   database, you should run this command.\n\n\
   Note that files added to the database using the 'addurl' e-command will\n\
   NOT be checked/updated in any way.\n\n\
   This command takes no parameters.\n\n\
EXAMPLE:\n\
      $ vitunes -e update\n\n\
");
}

void
ecmd_help_flush(void)
{  printf("\
VITUNES COMMAND:\n\tflush - dump output of vitunes database to terminal\n\n\
SYNOPSIS:\n\tflush\n\n\
DESCRIPTION:\n\
   The flush command simply outputs the contents of the meta information\n\
   database used by vitunes to stdout, in a fairly easy to read (but very\n\
   easy to parse/grep through) format.\n\n\
   This command is not intended to be used directly for any real purpose.\n\
   It is primarily provided for debugging purposes.\n\n\
   This command takes no parameters.\n\n\
EXAMPLE:\n\
   $ vitunes -e flush\n\n\
");
}

int
ecmd_help(int argc, char *argv[])
{
   bool found;
   int  i;

   if (argc > 2)
      errx(1, "usage: -e %s [command]", argv[0]);

   /* no help requested for a specific command, give a list of all e-cmds */
   if (argc == 1)
   {  printf("\
The following is a list of \"e-commands\" that are used up manipulate the\n\
media database that vitunes maintains.  Each command is executed by doing:\n\n\
   $ vitunes -e COMMAND [args ...]\n\n\
A more detailed explanation of each can be obtained by doing:\n\n\
   $ vitunes -e help COMMAND\n\n\
The list of available commands are:\n\n\
   COMMAND     BRIEF DESCRIPTION\n\
   -------     ------------------------------------------------------------\n\
   init        Create the initial files used by vitunes.\n\n\
   update      Load the existing database and check all media files\n\
               contained within it to see if any of they have been removed\n\
               or if their meta information has changed.  The library is\n\
               updated accordingly.\n\n\
   add         Scan the list of provided files/directories for files to add\n\
               to the database used by vitunes.\n\n\
   addurl      Add a URL to the database, where you provide your own meta\n\
               information.\n\n\
   check       Check an individiaul file to see its meta information and\n\
               whether or not it's in the database.\n\n\
   rmfile      Remove a file/URL from the database.\n\n\
   flush       Load the existing database and dump it's information in an\n\
               easy-to-parse format to stdout.\n\n\
   help        This command.\n\n\
");
      return 0;
   }

   /* if reach here, help fora specific command was requested */
   found = false;
   for (i = 0; i < ECMD_PATH_SIZE && !found; i++) {
      if (strcmp(argv[1], "help") == 0) {
         printf("You're a damn fool if you need help with help.\n");
         found = true;
      }
      else if (strcmp(argv[1], ECMD_PATH[i].name) == 0) {
         (ECMD_PATH[i].help)();
         found = true;
      }
   }

   if (!found)
      printf("Unknown command \"%s\".  See \"vitunes -e help\" for list.\n", argv[1]);

   return 0;
}

/* load config file and apply all changes within.
 * note that this requires mdb, ui, and player to all be loaded/setup
 */
void
load_config()
{
   size_t  length, linenum;
   bool    found;
   FILE   *fin;
   char   *line;
   char   *copy;
   char  **argv;
   int     argc;
   int     i;

   if ((fin = fopen(conf_file, "r")) == NULL)
      return;

   linenum = 0;
   while (!feof(fin)) {

      /* get next line */
      if ((line = fparseln(fin, &length, &linenum, NULL, 0)) == NULL) {
         if (ferror(fin))
            err(1, "error reading config file '%s'", conf_file);
         else
            break;
      }

      /* skip whitespace */
      copy = line;
      copy += strspn(copy, " \t\n");
      if (copy[0] == '\0') {
         free(line);
         continue;
      }

      /* parse line into argc/argv */
      switch (str2argv(copy, &argc, &argv)) {
         case PARGV_OK:
            break;
         case PARGV_ERR_ESCAPE:
            endwin();
            errx(1, "%s line %zd: unused escape sequence", conf_file, linenum);
            break;
         case PARGV_ERR_QUOTES:
            endwin();
            errx(1, "%s line %zd: mismatched quotes", conf_file, linenum);
            break;
         default:
            endwin();
            errx(1, "load_config: str2argv retuned unknown");
      }

      /* run commnad */
      found = false;
      for (i = 0; i < CommandPathSize; i++) {
         if (strcmp(argv[0], CommandPath[i].name) == 0) {
            found = true;
            if (((CommandPath[i].func)(argc, argv)) < 0) {
               endwin();
               errx(1, "%s line %zd: syntax error / bad command", conf_file, linenum);
            }
         }
      }

      if (!found) {
         endwin();
         errx(1, "%s line %zd: unknown command", conf_file, linenum);
      }

      argv_free(&argc, &argv);
      free(line);
   }

   fclose(fin);
}

/* parse the command line and handle all switches.
 * this also handles all of the e-commands.
 */
int
handle_switches(int argc, char *argv[])
{
   int ch;
   int i;

   while ((ch = getopt(argc, argv, "he:f:d:p:")) != -1) {
      switch (ch) {
         case 'e':
            argc -= optind - 1;
            argv += optind - 1;

            for (i = 0; i < ECMD_PATH_SIZE; i++) {
               if (strcmp(optarg, ECMD_PATH[i].name) == 0)
                  exit((ECMD_PATH[i].ecmd)(argc, argv));
            }

            errx(1, "Unknown e-command '%s'.  See 'vitunes -e help' for list.",
               optarg);
            break;

         case 'f':
            if ((conf_file = strdup(optarg)) == NULL)
               err(1, "failed to strdup -f conf-file");
            break;

         case 'd':
            if ((db_file = strdup(optarg)) == NULL)
               err(1, "failed to strdup -d db_file");
            break;

         case 'p':
            if ((playlist_dir = strdup(optarg)) == NULL)
               err(1, "failed to strdup -p playlist_dir");
            break;

         case 'h':
         case '?':
         default:
            usage(argv[0]);
            /* NOT REACHED */
      }
   }

   return 0;
}
